/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 24/07/22.
//

#include "fggl/entity/loader/loader.hpp"
#include "fggl/debug/logging.hpp"
#include "fggl/scenes/game.hpp"

namespace fggl::entity {

	assets::AssetRefRaw load_scene(assets::Loader* loader, const assets::AssetGUID& asset, assets::AssetData data, void* ptr) {
		auto *filePath = std::get<assets::AssetPath *>(data);
		scenes::Game* gamePtr = (scenes::Game*)ptr;

		// load assets
		auto* entityFactory = gamePtr->owner().service<EntityFactory>();

		auto nodes = YAML::LoadAllFromFile(filePath->c_str());
		for (const auto& node : nodes) {
			auto scene = node["scene"];
			if ( !scene ) {
				debug::warning("no scene node in YAML file...");
				return nullptr;
			}

			// create and personalize
			for (const auto& item : scene) {

				// only safe if personalize is called BEFORE end of loop itr
				auto personalize = [&entityFactory, &item, gamePtr](EntityManager& manager, const EntityID eid) {

					for ( const auto& compConfig : item["components"]) {
						auto compName = compConfig.first.as<fggl::util::GUID>();
						auto& compInfo = entityFactory->getInfo(compName);

						// setup config data
						ComponentSpec spec;
						spec.config = compConfig.second;

						// re-run the factory with the modified arguments (might need to support a 'patch' function...)
						compInfo.factory( spec, manager, eid, gamePtr->owner().services());
					}
				};

				// finally, load the entity and trigger personalize
				auto prefab = item["prefab"].as<fggl::util::GUID>();
				auto entity = entityFactory->create(prefab, gamePtr->world(), personalize);

				// metadata
				if ( item["name"].IsDefined() ) {
					gamePtr->world().setName(item["name"].as<std::string>(), entity);
				}

				if ( item["tags"].IsDefined() ) {
					for ( auto tag : item["tags"] ) {
						auto tagGuid = tag.as<fggl::util::GUID>();
						gamePtr->world().addTag(entity, tagGuid);
					}
				}
			}
		}

		return nullptr;
	}

	assets::AssetRefRaw load_prototype(assets::Loader* loader, const assets::AssetGUID &guid, assets::AssetData data, EntityFactory* factory) {
		auto *filePath = std::get<assets::AssetPath *>(data);

		// We need to process the prototypes, and load them into the asset system.
		auto nodes = YAML::LoadAllFromFile(filePath->c_str());
		for (const auto &node : nodes) {
			auto prefabs = node["prefabs"];

			for (const auto &prefab : prefabs) {
				auto name = prefab["name"].as<fggl::util::GUID>();
				debug::info("found prefab: {}", name);

				// set up the components
				EntitySpec entity{};
				entity.parent = prefab["parent"].as<fggl::util::GUID>(NO_PARENT);

				for (const auto &compEntry : prefab["components"]) {
					auto compId = compEntry.first.as<fggl::util::GUID>();

					ComponentSpec compSpec{};
					compSpec.config = compEntry.second;
					entity.components[compId] = compSpec;
					entity.ordering.push_back(compId);

					debug::trace("prefab {} has component {}", name, compId);
				}

				if ( prefab["tags"].IsDefined() ) {
					for ( auto& tagNode : prefab["tags"] ) {
						entity.tags.push_back( tagNode.as< util::GUID >() );
					}
				}

				factory->define(name, entity);
			}
		}

		return nullptr;
	}

} // namespace fggl::entity